import socket
import time
from enum import Enum
from threading import Thread
import random

from ingenialink.constants import ETH_BUF_SIZE, MONITORING_BUFFER_SIZE
from ingenialink.utils.mcb import MCB
from ingenialink.utils._utils import convert_bytes_to_dtype, convert_dtype_to_bytes
from ingenialink.ethernet.servo import EthernetServo
from ingenialink.ethernet.dictionary import EthernetDictionary
from ingenialink.enums.register import REG_DTYPE, REG_ACCESS
from ingenialink.utils import constants


class MSG_TYPE(Enum):
    RECEIVED = "RECEIVED"
    SENT = "SENT"


class VirtualMonDistBase:
    """Base class to implement VirtualMonitoring and VirtualDisturbance.

    Args:
        drive (VirtualDrive): Virtual drive instance.

    """

    FREQUENCY = 20000
    FREQ_DIVIDER_REG = None
    BUFFER_SIZE_REG = None
    NUMBER_MAP_REGS = None
    MAP_REG_CFG = None
    BYTES_PER_BLOCK_REG = None
    AVAILABLE_BYTES_REG = None
    DATA_REG = None

    def __init__(self, drive):
        self.drive = drive
        self.enabled = False
        self.channels = {}
        self.number_mapped_registers = 0

    def enable(self):
        """Enable Monitoring/Disturbance."""
        self.enabled = True

    def disable(self):
        """Disable Monitoring/Disturbance."""
        self.enabled = False

    def remove_data(self):
        """Remove Monitoring/Disturbance data."""
        for channel in range(self.number_mapped_registers):
            self.channels[channel]["data"] = []
            self.drive.set_value_by_id(0, self.DATA_REG, 0)

    @property
    def divider(self):
        """int: Frequency divider."""
        return self.drive.get_value_by_id(0, self.FREQ_DIVIDER_REG)

    @property
    def buffer_size(self):
        """int: Monitoring buffer size."""
        return self.drive.get_value_by_id(0, self.BUFFER_SIZE_REG)

    @property
    def bytes_per_block(self):
        """int: Monitoring bytes per sample."""
        return self.drive.get_value_by_id(0, self.BYTES_PER_BLOCK_REG)

    @bytes_per_block.setter
    def bytes_per_block(self, n_bytes):
        self.drive.set_value_by_id(0, self.BYTES_PER_BLOCK_REG, n_bytes)

    @property
    def number_mapped_registers(self):
        """int: Number of mapped registers."""
        return self.drive.get_value_by_id(0, self.NUMBER_MAP_REGS)

    @number_mapped_registers.setter
    def number_mapped_registers(self, value):
        return self.drive.set_value_by_id(0, self.NUMBER_MAP_REGS, value)

    @property
    def available_bytes(self):
        """int: Actual number of monitoring bytes."""
        return self.drive.get_value_by_id(0, self.AVAILABLE_BYTES_REG)

    @available_bytes.setter
    def available_bytes(self, n_bytes):
        self.drive.set_value_by_id(0, self.AVAILABLE_BYTES_REG, n_bytes)

    def get_mapped_register(self, channel):
        """Decodes the register with the information of a mapped register.

        Args:
            channel (int): Channel of the register to be decoded.

        Returns:
            int: Register subnode
            int: Register address
            int: Register dtype index
            int: Channel size

        """
        register_id = self.MAP_REG_CFG.format(channel)
        data = self.drive.get_value_by_id(0, register_id)
        data_h = data >> 16
        data_l = data & 0x0000FFFF
        subnode = data_h >> 12
        address = data_h & 0x0FFF
        dtype = data_l >> 8
        size = data_l & 0x00FF
        return subnode, address, dtype, size

    def map_registers(self):
        """Creates the channels attribute based on mapped registers."""
        self.bytes_per_block = 0
        for channel in range(self.number_mapped_registers):
            subnode, address, dtype, size = self.get_mapped_register(channel)
            self.channels[channel] = {
                "data": [],
                "dtype": REG_DTYPE(dtype),
                "address": address,
                "subnode": subnode,
                "size": size,
                "signal": [],
            }
            self.bytes_per_block += size


class VirtualMonitoring(VirtualMonDistBase):
    """Emulates monitoring at the VirtualDrive.

    Args:
        drive (VirtualDrive): Virtual drive instance.

    """

    FREQ_DIVIDER_REG = "MON_DIST_FREQ_DIV"
    BUFFER_SIZE_REG = "MON_CFG_WINDOW_SAMP"
    NUMBER_MAP_REGS = "MON_CFG_TOTAL_MAP"
    MAP_REG_CFG = "MON_CFG_REG{}_MAP"
    BYTES_PER_BLOCK_REG = "MON_CFG_BYTES_PER_BLOCK"
    DATA_REG = "MON_DATA"
    TRIGGER_TYPE_REG = "MON_CFG_SOC_TYPE"
    AVAILABLE_BYTES_REG = "MON_CFG_BYTES_VALUE"

    def __init__(self, drive):
        self.start_time = None
        super().__init__(drive)

    def enable(self):
        super().map_registers()
        super().enable()
        self.__create_signals()

    def disable(self):
        if self.enabled is False or self.start_time is None:
            return
        sampling_rate = self.FREQUENCY / self.divider
        elapsed_time = time.time() - self.start_time
        elapsed_samples = int(elapsed_time * sampling_rate)
        n_samples = min(elapsed_samples, self.buffer_size)
        for channel in range(self.number_mapped_registers):
            self.channels[channel]["data"] = self.channels[channel]["signal"][:n_samples]
        self.available_bytes = n_samples * self.bytes_per_block
        self._store_data_bytes()
        self.start_time = None
        super().disable()

    def trigger(self):
        """Triggers monitoring."""
        if self.enabled:
            self.start_time = time.time()

    def __create_signals(self):
        """Creates emulated monitoring signals."""
        for channel in range(self.number_mapped_registers):
            start_value = self.channels[channel]["address"] + self.channels[channel]["subnode"]
            signal = [
                start_value + i for i in range(0, self.buffer_size * self.divider, self.divider)
            ]
            self.channels[channel]["signal"] = signal

    def _store_data_bytes(self):
        """Convert signals into a bytearray and store it at MON_DATA register."""
        bytes = bytearray()
        n_samples = len(self.channels[0]["data"])
        for sample in range(n_samples):
            for channel in range(self.number_mapped_registers):
                value = self.channels[channel]["data"][sample]
                size = self.channels[channel]["size"]
                sample_bytes = convert_dtype_to_bytes(value, self.channels[channel]["dtype"])
                if len(sample_bytes) < size:
                    sample_bytes += b"0" * (size - len(sample_bytes))
                bytes += sample_bytes
        self.drive.set_value_by_id(0, "MON_DATA", bytes)

    @property
    def trigger_type(self):
        """int: Trigger type Auto(0), Force (1) or Rising or Failing (2)."""
        return self.drive.get_value_by_id(0, self.TRIGGER_TYPE_REG)


class VirtualDisturbance(VirtualMonDistBase):
    """Emulates disturbance at the VirtualDrive.

    Args:
        drive (VirtualDrive): Virtual drive instance.

    """

    FREQ_DIVIDER_REG = "DIST_FREQ_DIV"
    BUFFER_SIZE_REG = "DIST_CFG_SAMPLES"
    NUMBER_MAP_REGS = "DIST_CFG_MAP_REGS"
    MAP_REG_CFG = "DIST_CFG_REG{}_MAP"
    BYTES_PER_BLOCK_REG = "DIST_CFG_BYTES_PER_BLOCK"
    AVAILABLE_BYTES_REG = "DIST_CFG_BYTES"
    DATA_REG = "DIST_DATA"

    def __init__(self, drive):
        self.start_time = None
        self.received_bytes = bytearray()
        super().__init__(drive)

    def enable(self):
        super().enable()

    def append_data(self, data):
        """Append received disturbance data until the buffer is full.

        Args:
            data (bytearray): Received data.
        """
        if len(self.channels) == 0:
            super().map_registers()
        self.received_bytes += data
        self.available_bytes = len(self.received_bytes)
        if self.available_bytes == self.buffer_size_bytes:
            self.write_data()

    def write_data(self):
        """Convert received data and store it at the channels attribute."""
        n_samples = self.buffer_size
        buffer = self.received_bytes
        for sample in range(n_samples):
            for channel in range(self.number_mapped_registers):
                size = self.channels[channel]["size"]
                dtype = self.channels[channel]["dtype"]
                bytes = buffer[:size]
                value = convert_bytes_to_dtype(bytes, dtype)
                self.channels[channel]["data"].append(value)
                buffer = buffer[size:]

    @property
    def buffer_size_bytes(self):
        """int: Buffer size in bytes."""
        buffer_size_bytes = 0
        n_samples = self.buffer_size
        for channel in range(self.number_mapped_registers):
            buffer_size_bytes += self.channels[channel]["size"] * n_samples
        return buffer_size_bytes


class VirtualDrive(Thread):
    """Emulates a drive by creating a UDP server that sends and receives MCB messages.

    Args:
        ip (str): Server IP address.
        port (int): Server port number.
        dictionary_path (str): Path to the dictionary.

    """

    ACK_CMD = 3
    WRITE_CMD = 2
    READ_CMD = 1

    def __init__(self, ip, port, dictionary_path="./tests/resources/virtual_drive.xdf"):
        super(VirtualDrive, self).__init__()
        self.ip = ip
        self.port = port
        self.dictionary_path = dictionary_path
        self.socket = None
        self.__stop = False
        self.device_info = None
        self.__logger = []
        self.__reg_address_to_id = {}
        self.__dictionary = EthernetDictionary(dictionary_path)
        self._init_registers()
        self._update_registers()
        self.__monitoring = VirtualMonitoring(self)
        self.__disturbance = VirtualDisturbance(self)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    def run(self):
        """Open socket, listen and decode messages."""
        server_address = (self.ip, self.port)
        self.socket.bind(server_address)
        self.socket.settimeout(2)
        while not self.__stop:
            try:
                frame, add = self.socket.recvfrom(ETH_BUF_SIZE)
            except:
                continue
            reg_add, subnode, cmd, data = MCB.read_mcb_frame(frame)
            self.__log(add, frame, MSG_TYPE.RECEIVED)
            register = self.get_register(subnode, reg_add)
            if cmd == self.WRITE_CMD:
                sent_cmd = self.ACK_CMD
                response = MCB.build_mcb_frame(sent_cmd, subnode, reg_add, data[:8])
                if register.access in [REG_ACCESS.RW, REG_ACCESS.WO]:  # TODO: send error otherwise
                    value = convert_bytes_to_dtype(data, register.dtype)
                    self.set_value_by_id(subnode, register.identifier, value)
                    self.__decode_msg(reg_add, subnode, data)
            elif cmd == self.READ_CMD:
                value = self.get_value_by_id(subnode, register.identifier)
                data = convert_dtype_to_bytes(value, register.dtype)
                sent_cmd = self.ACK_CMD
                if reg_add == self.id_to_address(0, "MON_DATA"):
                    response = self._response_monitoring_data(data)
                else:
                    response = MCB.build_mcb_frame(sent_cmd, subnode, reg_add, data)
                # TODO: send error if the register is WO
            else:
                continue
            self.__send(response, add)

        time.sleep(0.1)

    def stop(self):
        """Stop socket."""
        if self.socket is not None:
            self.socket.close()
        self.__stop = True
        self.__monitoring.disable()

    def _init_registers(self):
        """Initialize some relevant registers."""
        self.set_value_by_id(0, "DRV_ID_PRODUCT_CODE_COCO", 123456)
        self.set_value_by_id(1, "DRV_ID_PRODUCT_CODE", 123456)
        self.set_value_by_id(0, "DRV_ID_REVISION_NUMBER_COCO", 654321)
        self.set_value_by_id(1, "DRV_ID_REVISION_NUMBER", 654321)
        self.set_value_by_id(0, "DRV_APP_COCO_VERSION", "0.1.0")
        self.set_value_by_id(1, "DRV_ID_SOFTWARE_VERSION", "0.1.0")
        self.set_value_by_id(1, "DRV_ID_SOFTWARE_VERSION", "0.1.0")
        self.set_value_by_id(1, "DRV_STATE_STATUS", constants.IL_MC_PDS_STA_RTSO)
        self.set_value_by_id(1, "DRV_POS_VEL_RATE", 20000)
        self.set_value_by_id(0, "DIST_MAX_SIZE", 8192)

    def _update_registers(self):
        """Force storage_valid at each register and add registers that are not in the dictionary."""
        for subnode in range(self.__dictionary.subnodes):
            self.__reg_address_to_id[subnode] = {}
            for reg_id, reg in self.__dictionary.registers(subnode).items():
                self.__reg_address_to_id[subnode][reg.address] = reg_id
                self.__dictionary.registers(subnode)[reg_id].storage_valid = 1

        custom_regs = {
            "MON_DATA": EthernetServo.MONITORING_DATA,
            "DIST_DATA": EthernetServo.DIST_DATA,
        }
        for id, reg in custom_regs.items():
            register = {
                "address": reg.address,
                "access": reg.access,
                "dtype": REG_DTYPE.DOMAIN,
                "identifier": id,
                "subnode": reg.subnode,
            }
            self.__dictionary._add_register_list(register)
            self.__dictionary.registers(reg.subnode)[id].storage_valid = 1

            self.__reg_address_to_id[reg.subnode][reg.address] = id

    def __send(self, response, address):
        """Send a message and update log.

        Args:
            response (bytes): Message to be sent.
            address (tuple): IP address and port.
        """
        time.sleep(0.01)  # Emulate latency of 10 ms
        self.socket.sendto(response, address)
        self.__log(address, response, MSG_TYPE.SENT)

    def _response_monitoring_data(self, data):
        """Creates a response for monitoring data.

        Args:
            data (bytes): Data to be sent.

        Returns:
            bytes: MCB frame.
        """
        sent_cmd = self.ACK_CMD
        reg_add = self.id_to_address(0, "MON_DATA")
        limit = min(len(data), MONITORING_BUFFER_SIZE)
        response = MCB.build_mcb_frame(sent_cmd, 0, reg_add, data[:limit])
        data_left = data[limit:]
        self.set_value_by_id(0, "MON_DATA", data_left)
        self.__monitoring.available_bytes = len(data_left)
        return response

    def __log(self, ip_port, message, msg_type):
        """Updates log.

        Args:
            ip_port (tuple): IP address and port.
            message (bytes): Received or sent message.
            msg_type (MSG_TYPE): Sent or Received.
        """
        self.__logger.append(
            {
                "timestamp": time.time(),
                "ip_port": ip_port,
                "type": msg_type.value,
                "message": message,
            }
        )

    @property
    def log(self):
        """dict: Dictionary containing log information."""
        return self.__logger

    def clean_log(self):
        """Cleans log."""
        self.__logger = []

    def __decode_msg(self, reg_add, subnode, data):
        """Decodes received messages and run specific methods if needed.

        Args:
            reg_add (int): Register address.
            subnode (int): Subnode.
            data (bytes): Received data.
        """
        register = self.get_register(subnode, reg_add)
        reg_id = register.identifier
        dtype = register.dtype
        value = convert_bytes_to_dtype(data, dtype)
        if reg_id == "MON_DIST_ENABLE" and subnode == 0 and value == 1:
            self.__monitoring.enable()
        if reg_id == "MON_DIST_ENABLE" and subnode == 0 and value == 0:
            self.__monitoring.disable()
        if reg_id == "MON_CMD_FORCE_TRIGGER" and subnode == 0 and value == 1:
            self.__monitoring.trigger()
        if reg_id == "MON_REMOVE_DATA" and subnode == 0 and value == 1:
            self.__monitoring.remove_data()
        if reg_id == "DIST_ENABLE" and subnode == 0 and value == 1:
            self.__disturbance.enable()
        if reg_id == "DIST_ENABLE" and subnode == 0 and value == 0:
            self.__disturbance.disable()
        if reg_id == "DIST_REMOVE_DATA" and subnode == 0 and value == 1:
            self.__disturbance.remove_data()
        if reg_id == "DIST_DATA" and subnode == 0:
            self.__disturbance.append_data(data)

    def address_to_id(self, subnode, address):
        """Converts a register address into its ID.

        Args:
            subnode (int): Subnode.
            address (int): Register address.

        Returns:
            str: Register ID.
        """
        return self.__reg_address_to_id[subnode][address]

    def id_to_address(self, subnode, id):
        """Converts a register address into an ID.

        Args:
            subnode (int): Subnode.
            id (str): Register ID.

        Returns:
            int: Register adress.
        """
        register = self.__dictionary.registers(subnode)[id]
        return register.address

    def get_value_by_id(self, subnode, id):
        """Returns a register value by its ID.

        Args:
            subnode (int): Subnode.
            id (str): Register ID.

        Returns:
            (float, int, str): Register value.
        """
        register = self.__dictionary.registers(subnode)[id]
        if register._storage is None:
            range_value = register.range
            if not register.range[0]:
                range_value = (1, 10)
            value = random.uniform(*range_value)
            if register.dtype != REG_DTYPE.FLOAT:
                value = int(value)
                if register.dtype == REG_DTYPE.STR:
                    value = ""
            self.set_value_by_id(subnode, id, value)
        return self.__dictionary.registers(subnode)[id]._storage

    def set_value_by_id(self, subnode, id, value):
        """Set a register value by its ID.

        Args:
            subnode (int): Subnode.
            id (str): Register ID.
            value (int, float, str): Value to be set.
        """
        self.__dictionary.registers(subnode)[id].storage = value

    def get_register(self, subnode, address=None, id=None):
        """Returns a register by its address or ID.

        Args:
            subnode (int): Subnode.
            address (int): Register address. Default to None.
            id (str): Register ID. Default to None.

        Returns:
            EthernetRegister: Register instance.

        Raises:
            ValueError: If both address and id are None.
        """
        if address is not None:
            id = self.address_to_id(subnode, address)
        else:
            if id is None:
                raise ValueError("Register address or id should be passed")
        return self.__dictionary.registers(subnode)[id]



import sys

from ingenialink.canopen.network import CanopenNetwork, CAN_DEVICE, CAN_BAUDRATE


def connect_slave():
    net = CanopenNetwork(device=CAN_DEVICE.IXXAT,
                         channel=0,
                         baudrate=CAN_BAUDRATE.Baudrate_1M)

    servo = net.connect_to_slave(
        target=32,
        dictionary='../../resources/dictionaries/eve-net-c_can_1.8.1.xdf')
    
    return servo, net


def load_config_example():
    """Loads a given configuration file into the drive."""
    servo, net = connect_slave()
    servo.load_configuration(
        'can_config.xcf')
    servo.load_configuration(
        'can_config_0.xcf',
        subnode=0)
    servo.load_configuration(
        'can_config_1.xcf',
        subnode=1)

    net.disconnect_from_slave(servo)


def save_config_example():
    """Saves the drive configuration into a file."""

    servo, net = connect_slave()
    servo.save_configuration(
        'can_config.xcf')
    servo.save_configuration(
        'can_config_0.xcf',
        subnode=0)
    servo.save_configuration(
        'can_config_1.xcf',
        subnode=1)

    net.disconnect_from_slave(servo)


if __name__ == '__main__':
    save_config_example()
    load_config_example()
    sys.exit()



import ftplib
import time
from collections import defaultdict

from .servo import EthernetServo
from ingenialink.utils.udp import UDP
from ..network import NET_PROT
from ingenialink.network import Network, NET_STATE, NET_DEV_EVT
from ingenialink.exceptions import ILFirmwareLoadError, ILError
from ingenialink.constants import DEFAULT_ETH_CONNECTION_TIMEOUT

from ftplib import FTP
from time import sleep

import os
import socket
import ingenialogger
from threading import Thread

from ping3 import ping

logger = ingenialogger.get_logger(__name__)

FTP_SESSION_OK_CODE = "220"
FTP_LOGIN_OK_CODE = "230"
FTP_FILE_TRANSFER_OK_CODE = "226"
FTP_CLOSE_OK_CODE = "221"

CMD_CHANGE_CPU = 0x67E4

MAX_NUM_UNSUCCESSFUL_PINGS = 3


class NetStatusListener(Thread):
    """Network status listener thread to check if the drive is alive.

    Args:
        network (EthernetNetwork): Network instance of the Ethernet communication.

    """

    def __init__(self, network, refresh_time=0.25):
        super(NetStatusListener, self).__init__()
        self.__network = network
        self.__refresh_time = refresh_time
        self.__stop = False
        self.__max_unsuccessful_pings = MAX_NUM_UNSUCCESSFUL_PINGS

    def run(self):
        while not self.__stop:
            for servo in self.__network.servos:
                unsuccessful_pings = 0
                servo_ip = servo.ip_address
                servo_state = self.__network._get_servo_state(servo_ip)
                while unsuccessful_pings < self.__max_unsuccessful_pings:
                    response = servo.is_alive()  # TODO: Use ping after CAP-924 is fixed
                    if response == False:
                        unsuccessful_pings += 1
                    else:
                        break
                ping_response = unsuccessful_pings != self.__max_unsuccessful_pings
                if servo_state == NET_STATE.CONNECTED and not ping_response:
                    self.__network._notify_status(servo_ip, NET_DEV_EVT.REMOVED)
                    self.__network._set_servo_state(servo_ip, NET_STATE.DISCONNECTED)
                if servo_state == NET_STATE.DISCONNECTED and ping_response:
                    self.__network._notify_status(servo_ip, NET_DEV_EVT.ADDED)
                    self.__network._set_servo_state(servo_ip, NET_STATE.CONNECTED)
            time.sleep(self.__refresh_time)

    def stop(self):
        self.__stop = True


class EthernetNetwork(Network):
    """Network for all Ethernet communications."""

    def __init__(self):
        super(EthernetNetwork, self).__init__()
        self.__servos_state = {}
        self.__listener_net_status = None
        self.__observers_net_state = defaultdict(list)

    @staticmethod
    def load_firmware(fw_file, target="192.168.2.22", ftp_user="", ftp_pwd=""):
        """Loads a given firmware file to the target slave.

        .. warning ::
            It is needed to disconnect the drive(:func:`disconnect_from_slave`)
            after loading the firmware since the `Servo` object's data will
            become obsolete.

        Args:
            fw_file (str): Path to the firmware file to be loaded.
            target (str): IP of the target slave.
            ftp_user (str): FTP user to connect with.
            ftp_pwd (str): FTP password for the given user.

        Raises:
            ILError: If the loading firmware process fails.

        """
        if not os.path.isfile(fw_file):
            raise FileNotFoundError(f"Could not find {fw_file}.")

        try:
            file = open(fw_file, "rb")
            ftp_output = None
            ftp = FTP()

            # Start a FTP session. Drive must be in BOOT mode.
            logger.info("Starting FTP session...")
            ftp_output = ftp.connect(target)
            logger.info(ftp_output)
            if FTP_SESSION_OK_CODE not in ftp_output:
                raise ILError("Unable to open FTP session")

            # Login into FTP session.
            logger.info("Logging into FTP session...")
            ftp_output = ftp.login(ftp_user, ftp_pwd)
            logger.info(ftp_output)
            if FTP_LOGIN_OK_CODE not in ftp_output:
                raise ILError("Unable to login the FTP session")

            # Load file through FTP.
            logger.info("Uploading firmware file...")
            ftp.set_pasv(False)
            ftp_output = ftp.storbinary(f"STOR {os.path.basename(file.name)}", file)
            logger.info(ftp_output)
            if FTP_FILE_TRANSFER_OK_CODE not in ftp_output:
                raise ILError("Unable to load the FW file through FTP")

            # Close FTP session.
            logger.info("Closing FTP session...")
            ftp.close()

            # Close the temporal file
            file.close()

        except Exception as e:
            logger.error(e)
            raise ILFirmwareLoadError("Error during bootloader process.")

    @staticmethod
    def load_firmware_moco(node, subnode, ip, port, moco_file):
        """Update MOCO firmware through UDP protocol.

        Args:
            node: Network node.
            subnode: Drive subnode.
            ip: Drive address IP.
            port: Drive port.
            moco_file: Path to the firmware file.

        Returns:
            int: Result code.

        Raises:
            ILFirmwareLoadError: The firmware load process fails
                with an error message.
        """
        r = 0
        upd = UDP(port, ip)

        if not moco_file or not os.path.isfile(moco_file):
            raise ILFirmwareLoadError("File not found")
        moco_in = open(moco_file, "r")

        logger.info("Loading firmware...")
        try:
            for line in moco_in:
                words = line.split()

                # Get command and address
                cmd = int(words[1] + words[0], 16)
                data = b""
                data_start_byte = 2
                while data_start_byte in range(data_start_byte, len(words)):
                    # Load UDP data
                    data += bytes([int(words[data_start_byte], 16)])
                    data_start_byte += 1

                # Send message
                upd.raw_cmd(node, subnode, cmd, data)

                if cmd == CMD_CHANGE_CPU:
                    sleep(1)

            logger.info("Bootload process succeeded")
        except ftplib.error_temp as e:
            logger.error(e)
            raise ILFirmwareLoadError("Firewall might be blocking the access.")
        except Exception as e:
            logger.error(e)
            raise ILFirmwareLoadError("Error during bootloader process.")

    def scan_slaves(self):
        raise NotImplementedError

    def connect_to_slave(
        self,
        target,
        dictionary=None,
        port=1061,
        connection_timeout=DEFAULT_ETH_CONNECTION_TIMEOUT,
        servo_status_listener=False,
        net_status_listener=False,
    ):
        """Connects to a slave through the given network settings.

        Args:
            target (str): IP of the target slave.
            dictionary (str): Path to the target dictionary file.
            port (int): Port to connect to the slave.
            connection_timeout (float): Time in seconds of the connection timeout.
            servo_status_listener (bool): Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener (bool): Toggle the listener of the network
                status, connection and disconnection.

        Returns:
            EthernetServo: Instance of the servo connected.

        """
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.settimeout(connection_timeout)
        sock.connect((target, port))
        servo = EthernetServo(sock, dictionary, servo_status_listener)
        try:
            servo.get_state()
        except ILError as e:
            servo.stop_status_listener()
            raise ILError(f"Drive not found in IP {target}.") from e
        self.servos.append(servo)
        self._set_servo_state(target, NET_STATE.CONNECTED)

        if net_status_listener:
            self.start_status_listener()
        else:
            self.stop_status_listener()

        return servo

    def disconnect_from_slave(self, servo):
        """Disconnects the slave from the network.

        Args:
            servo (EthernetServo): Instance of the servo connected.

        """
        self.servos.remove(servo)
        servo.stop_status_listener()
        self.close_socket(servo.socket)
        self._set_servo_state(servo.ip_address, NET_STATE.DISCONNECTED)
        if len(self.servos) == 0:
            self.stop_status_listener()

    @staticmethod
    def close_socket(sock):
        """Closes the established network socket."""
        sock.shutdown(socket.SHUT_RDWR)
        sock.close()

    def start_status_listener(self):
        """Start monitoring network events (CONNECTION/DISCONNECTION)."""
        if self.__listener_net_status is None:
            listener = NetStatusListener(self)
            listener.start()
            self.__listener_net_status = listener

    def stop_status_listener(self):
        """Stops the NetStatusListener from listening to the drive."""
        if self.__listener_net_status is not None:
            self.__listener_net_status.stop()
            self.__listener_net_status.join()
        self.__listener_net_status = None

    def _notify_status(self, ip, status):
        """Notify subscribers of a network state change."""
        for callback in self.__observers_net_state[ip]:
            callback(status)

    def subscribe_to_status(self, ip, callback):
        """Subscribe to network state changes.

        Args:
            ip (str): IP of the drive to subscribe.
            callback (function): Callback function.

        """
        if callback in self.__observers_net_state[ip]:
            logger.info("Callback already subscribed.")
            return
        self.__observers_net_state[ip].append(callback)

    def unsubscribe_from_status(self, ip, callback):
        """Unsubscribe from network state changes.

        Args:
            ip (str): IP of the drive to unsubscribe.
            callback (function): Callback function.

        """
        if callback not in self.__observers_net_state[ip]:
            logger.info("Callback not subscribed.")
            return
        self.__observers_net_state[ip].remove(callback)

    def _get_servo_state(self, ip):
        return self.__servos_state[ip]

    def _set_servo_state(self, ip, state):
        self.__servos_state[ip] = state

    @property
    def protocol(self):
        """NET_PROT: Obtain network protocol."""
        return NET_PROT.ETH



#!/usr/bin/env python

import re
import setuptools

_version = re.search(r'__version__\s+=\s+\"(.*)\"',
                     open('ingenialink/__init__.py').read()).group(1)


def get_docs_url():
    return f"https://distext.ingeniamc.com/doc/ingenialink-python/{_version}"


setuptools.setup(
    name='ingenialink',
    version=_version,
    packages=setuptools.find_packages(exclude=["test", "examples"]),
    include_package_data=True,
    package_data={"ingenialink": ["bin/FoE/*/*"]},
    description='IngeniaLink Communications Library',
    long_description=open('README.rst').read(),
    author='Ingenia Motion Control',
    author_email='support@ingeniamc.com',
    url='https://www.ingeniamc.com',
    project_urls={
      'Documentation': get_docs_url(),
      'Source': 'https://github.com/ingeniamc/ingenialink-python'
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Communications',
        'Topic :: Software Development :: Libraries'
    ],
    install_requires=[
        'canopen==1.2.1',
        'python-can==3.3.4',
        'ingenialogger>=0.2.1',
        'ping3==4.0.3'
    ],
    extras_require={
        "dev": [
            "sphinx==3.5.4",
            "sphinx-rtd-theme==1.0.0",
            "sphinxcontrib-bibtex==2.4.1",
            "nbsphinx==0.8.6",
            "pytest==6.2.4",
            "pytest-cov==2.12.1",
            "pytest-mock==3.6.1",
            "jinja2==3.0.3",
            "pycodestyle==2.6.0",
            "wheel==0.37.1",
            "m2r2==0.3.2"
        ],
    },
)



import sys

from ingenialink.ethernet.network import EthernetNetwork, NET_TRANS_PROT


def eth_store_parameters():
    net = EthernetNetwork()
    servo = net.connect_to_slave("192.168.2.22",
                                 "../../resources/dictionaries/eve-net-c_eth_1.8.1.xdf",
                                 1061,
                                 NET_TRANS_PROT.UDP)

    print(servo.read('DRV_ID_SOFTWARE_VERSION'))

    try:
        servo.store_parameters(subnode=1)
        print('Store successful')
    except Exception as e:
        print('Error storing', e)

    net.disconnect_from_slave(servo)


def eth_restore_parameters():
    net = EthernetNetwork()
    servo = net.connect_to_slave("192.168.2.22",
                                 "../../resources/dictionaries/eve-net-c_eth_1.8.1.xdf",
                                 1061,
                                 NET_TRANS_PROT.UDP)

    print(servo.read('DRV_ID_SOFTWARE_VERSION'))

    try:
        servo.restore_parameters()
        print('Restore successful')
    except Exception as e:
        print('Error restoring', e)

    net.disconnect_from_slave(servo)


if __name__ == '__main__':
    eth_store_parameters()
    eth_restore_parameters()
    sys.exit(0)



import sys

from ingenialink.ethernet.network import EthernetNetwork, NET_TRANS_PROT


def connection_example():
    net = EthernetNetwork()
    servo = net.connect_to_slave("192.168.2.22",
                                 "../../resources/dictionaries/eve-net-c_eth_1.8.1.xdf",
                                 1061,
                                 NET_TRANS_PROT.UDP)

    print(servo.read('DRV_ID_SOFTWARE_VERSION'))

    net.disconnect_from_slave(servo)


if __name__ == '__main__':
    connection_example()
    sys.exit(0)



from enum import Enum


class REG_DTYPE(Enum):
    """Data Type."""

    U8 = 0
    """Unsigned 8-bit integer."""
    S8 = 1
    """Signed 8-bit integer."""
    U16 = 2
    """Unsigned 16-bit integer."""
    S16 = 3
    """Signed 16-bit integer."""
    U32 = 4
    """Unsigned 32-bit integer."""
    S32 = 5
    """Signed 32-bit integer."""
    U64 = 6
    """Unsigned 64-bit integer."""
    S64 = 7
    """Signed 64-bit integer."""
    FLOAT = 8
    """Float."""
    STR = 10
    """String."""
    DOMAIN = 15
    """Domain."""


class REG_ACCESS(Enum):
    """Access Type."""

    RW = 0
    """Read/Write."""
    RO = 1
    """Read-only."""
    WO = 2
    """Write-only."""


class REG_PHY(Enum):
    """Physical Units."""

    NONE = 0
    """None."""
    TORQUE = 1
    """Torque."""
    POS = 2
    """Position."""
    VEL = 3
    """Velocity."""
    ACC = 4
    """Acceleration."""
    VOLT_REL = 5
    """Relative voltage (DC)."""
    RAD = 6
    """Radians."""


class REG_ADDRESS_TYPE(Enum):
    """Address Type."""

    NVM = 0
    NVM_NONE = 1
    NVM_CFG = 2
    NVM_LOCK = 3
    NVM_HW = 4



from ingenialink.register import Register
from ingenialink.enums.register import REG_DTYPE, REG_ACCESS, REG_PHY, REG_ADDRESS_TYPE


class CanopenRegister(Register):
    """CANopen Register.

    Args:
        identifier (str): Identifier.
        units (str): Units.
        cyclic (str): Cyclic typed register.
        idx (int): Index of the register.
        subidx (int): Subindex of the register.
        dtype (REG_DTYPE): Data type.
        access (REG_ACCESS): Access type.
        phy (REG_PHY, optional): Physical units.
        subnode (int): Subnode.
        storage (any, optional): Storage.
        reg_range (tuple, optional): Range (min, max).
        labels (dict, optional): Register labels.
        enums (dict): Enumeration registers.
        enums_count (int): Number of enumeration registers.
        cat_id (str, optional): Category ID.
        scat_id (str, optional): Sub-category ID.
        internal_use (int, optional): Internal use.
        address_type (REG_ADDRESS_TYPE): Address Type.

    Raises:
        TypeError: If any of the parameters has invalid type.
        ILValueError: If the register is invalid.
        ILAccessError: Register with wrong access type.

    """

    def __init__(
        self,
        identifier,
        units,
        cyclic,
        idx,
        subidx,
        dtype,
        access,
        phy=REG_PHY.NONE,
        subnode=1,
        storage=None,
        reg_range=(None, None),
        labels=None,
        enums=None,
        cat_id=None,
        scat_id=None,
        internal_use=0,
        address_type=None,
    ):
        super().__init__(
            dtype,
            access,
            identifier,
            units,
            cyclic,
            phy,
            subnode,
            storage,
            reg_range,
            labels,
            enums,
            cat_id,
            scat_id,
            internal_use,
            address_type,
        )

        self.__idx = idx
        self.__subidx = subidx

    @property
    def idx(self):
        """int: Register index."""
        return self.__idx

    @property
    def subidx(self):
        """int: Register subindex."""
        return self.__subidx

    @property
    def mapped_address(self):
        """int: Register mapped address used for monitoring/disturbance."""
        return self.idx



import time
import socket

import pytest

from ingenialink.ethernet.network import EthernetNetwork, NET_PROT, NET_STATE, NET_DEV_EVT
from ingenialink.exceptions import ILFirmwareLoadError, ILError


@pytest.fixture()
def connect(read_config):
    net = EthernetNetwork()
    protocol_contents = read_config["ethernet"]
    servo = net.connect_to_slave(
        protocol_contents["ip"], protocol_contents["dictionary"], protocol_contents["port"]
    )
    return servo, net


@pytest.mark.ethernet
def test_connect_to_slave(connect_to_slave):
    servo, net = connect_to_slave
    assert servo is not None and net is not None
    assert len(net.servos) == 1
    fw_version = servo.read("DRV_ID_SOFTWARE_VERSION")
    assert fw_version is not None and fw_version != ""


@pytest.mark.ethernet
def test_can_not_connect_to_salve(read_config):
    net = EthernetNetwork()
    wrong_ip = "34.56.125.234"
    protocol_contents = read_config["ethernet"]
    with pytest.raises(ILError):
        net.connect_to_slave(wrong_ip, protocol_contents["dictionary"], protocol_contents["port"])


@pytest.mark.ethernet
def test_scan_slaves(read_config):
    # TODO: Not implemented
    # net = EthernetNetwork()
    # slaves = net.scan_slaves()
    # assert len(slaves) > 0
    pass


@pytest.mark.ethernet
def test_ethernet_connection(connect_to_slave, read_config):
    servo, net = connect_to_slave
    family = servo.socket.family
    ip, port = servo.socket.getpeername()
    assert net._get_servo_state(read_config["ethernet"]["ip"]) == NET_STATE.CONNECTED
    assert net.protocol == NET_PROT.ETH
    assert family == socket.AF_INET
    assert servo.socket.type == socket.SOCK_DGRAM
    assert ip == read_config["ethernet"]["ip"]
    assert port == read_config["ethernet"]["port"]


@pytest.mark.ethernet
def test_ethernet_disconnection(connect, read_config):
    servo, net = connect
    net.disconnect_from_slave(servo)
    assert net._get_servo_state(read_config["ethernet"]["ip"]) == NET_STATE.DISCONNECTED
    assert len(net.servos) == 0
    assert servo.socket._closed


@pytest.mark.no_connection
def test_load_firmware_file_not_found():
    virtual_net = EthernetNetwork()
    with pytest.raises(FileNotFoundError):
        virtual_net.load_firmware("no_file")


@pytest.mark.no_connection
def test_load_firmware_no_connection(read_config):
    protocol_contents = read_config["ethernet"]
    virtual_net = EthernetNetwork()
    with pytest.raises(ILFirmwareLoadError):
        virtual_net.load_firmware(
            protocol_contents["fw_file"], target="127.0.0.1", ftp_user="", ftp_pwd=""
        )


@pytest.mark.skip
@pytest.mark.no_connection
def test_load_firmware_wrong_user_pwd():
    # TODO: implement
    pass


@pytest.mark.skip
@pytest.mark.no_connection
def test_load_firmware_error_during_loading():
    # TODO: implement
    pass


@pytest.mark.no_connection
def test_net_status_listener_connection(virtual_drive, read_config):
    server = virtual_drive
    net = EthernetNetwork()
    protocol_contents = read_config["ethernet"]
    status_list = []
    net.connect_to_slave(server.ip, dictionary=protocol_contents["dictionary"], port=server.port)

    status_list = []
    net.subscribe_to_status(server.ip, status_list.append)
    # Emulate a disconnection. TODO: disconnect from the virtual drive
    net._set_servo_state(server.ip, NET_STATE.DISCONNECTED)
    net.start_status_listener()
    time.sleep(2)
    net.stop_status_listener()

    assert len(status_list) == 1
    assert status_list[0] == NET_DEV_EVT.ADDED


@pytest.mark.skip
@pytest.mark.no_connection
def test_net_status_listener_disconnection():
    # TODO: implement
    pass


@pytest.mark.no_connection
def test_unsubscribe_from_status(virtual_drive, read_config):
    server = virtual_drive
    net = EthernetNetwork()
    protocol_contents = read_config["ethernet"]

    status_list = []
    net.connect_to_slave(server.ip, dictionary=protocol_contents["dictionary"], port=server.port)

    status_list = []
    net.subscribe_to_status(server.ip, status_list.append)
    net.unsubscribe_from_status(server.ip, status_list.append)

    # Emulate a disconnection. TODO: disconnect from the virtual drive
    net._set_servo_state(server.ip, NET_STATE.DISCONNECTED)
    net.start_status_listener()
    time.sleep(2)
    net.stop_status_listener()

    assert len(status_list) == 0



import sys
from ingenialink.ethernet.network import EthernetNetwork


def eth_load_fw():
    net = EthernetNetwork()

    net.load_firmware(fw_file='../../resources/firmware/eve-net-c_1.8.1.sfu',
                      ftp_user='Ingenia', ftp_pwd='Ingenia')


if __name__ == '__main__':
    eth_load_fw()
    sys.exit(0)



import pytest

from canopen import Network

from ingenialink.canopen.network import CanopenNetwork, CAN_DEVICE, CAN_BAUDRATE, NET_STATE
from ingenialink.exceptions import ILError

test_bus = "virtual"
test_baudrate = 1000000
test_channel = 0


@pytest.fixture
def virtual_network():
    net = CanopenNetwork(
        device=CAN_DEVICE(test_bus), channel=test_channel, baudrate=CAN_BAUDRATE(test_baudrate)
    )
    return net


@pytest.mark.no_connection
def test_getters_canopen(virtual_network):
    assert virtual_network.device == test_bus
    assert virtual_network.channel == test_channel
    assert virtual_network.baudrate == test_baudrate
    assert virtual_network.network is None
    assert virtual_network.status == NET_STATE.DISCONNECTED


@pytest.mark.canopen
def test_connect_to_slave(connect_to_slave):
    servo, net = connect_to_slave
    assert servo is not None and net is not None
    assert len(net.servos) == 1
    fw_version = servo.read("DRV_ID_SOFTWARE_VERSION")
    assert fw_version is not None and fw_version != ""


@pytest.mark.canopen
def test_connect_to_slave_target_not_in_nodes(read_config):
    protocol_contents = read_config["canopen"]
    net = CanopenNetwork(
        device=CAN_DEVICE(protocol_contents["device"]),
        channel=protocol_contents["channel"],
        baudrate=CAN_BAUDRATE(protocol_contents["baudrate"]),
    )

    with pytest.raises(ILError):
        net.connect_to_slave(target=1234, dictionary=protocol_contents["dictionary"])
    net._teardown_connection()


@pytest.mark.no_connection
def test_connect_to_slave_none_nodes(virtual_network, read_config):
    net = virtual_network
    protocol_contents = read_config["canopen"]
    with pytest.raises(ILError):
        net.connect_to_slave(target=1, dictionary=protocol_contents["dictionary"])


@pytest.mark.canopen
def test_scan_slaves(read_config):
    net = CanopenNetwork(
        device=CAN_DEVICE(read_config["canopen"]["device"]),
        channel=read_config["canopen"]["channel"],
        baudrate=CAN_BAUDRATE(read_config["canopen"]["baudrate"]),
    )
    slaves = net.scan_slaves()
    assert len(slaves) > 0


@pytest.mark.no_connection
def test_scan_slaves_none_nodes(virtual_network):
    nodes = virtual_network.scan_slaves()
    assert len(nodes) == 0


@pytest.mark.canopen
def test_disconnect_from_slave(read_config):
    protocol_contents = read_config["canopen"]
    net = CanopenNetwork(
        device=CAN_DEVICE(protocol_contents["device"]),
        channel=protocol_contents["channel"],
        baudrate=CAN_BAUDRATE(protocol_contents["baudrate"]),
    )

    servo = net.connect_to_slave(
        target=protocol_contents["node_id"],
        dictionary=protocol_contents["dictionary"],
    )

    assert len(net.servos) == 1
    net.disconnect_from_slave(servo)
    assert len(net.servos) == 0


@pytest.mark.no_connection
def test_setup_and_teardown_connection(virtual_network):
    assert virtual_network._connection is None
    virtual_network._setup_connection()
    assert virtual_network._connection is not None
    virtual_network._teardown_connection()
    assert virtual_network._connection is None


@pytest.mark.skip
def test_load_firmware(connect_to_slave, read_config):
    # TODO: Fix load_firmware method to work independently of status listeners
    servo, net = connect_to_slave
    assert servo is not None and net is not None
    assert len(net.servos) == 1
    fw_version = servo.read("DRV_ID_SOFTWARE_VERSION")
    protocol_contents = read_config["canopen"]

    net.load_firmware(protocol_contents["node_id"], protocol_contents["test_fw_file"])
    new_fw_version = servo.read("DRV_ID_SOFTWARE_VERSION")

    assert new_fw_version != fw_version
    net.disconnect_from_slave(servo)
